#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <sys/socket.h>
#include <time.h>
#include <signal.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <math.h>

#include <vector>

#include "utils.h"

#include "tcp_sink.h"
#include "rawdnet.h"
#include "sendchain.h"
#include "ipfrag.h"
#include <err.h>
#include "dnet.h"
#include "netinet/in_systm.h"
#include "netinet/ip.h"
#include "netinet/ip_icmp.h"

/**********************************************************
 *	                                     Global Variables
 **********************************************************/

char g_rcv_packet[IP_LEN_MAX];		// Last inbound ICMP packet,  IP_LEN_MAX = 65535
int	g_rcv_icmp_sock;
int	g_rcv_tcp_sock;

char *g_prog = "IPFrag";				// program name
int g_count = 100;						// probe count
int g_freq = 10;						// frequncy of sending packets
double g_interval = 0.1;				// Interval between probes
double g_frag_interval = -1;			// interval between fragments of a probe
int g_pktlen = 100;					// packet length
char *g_source =  NULL;				// source ip address

int g_int_state = 0;					// state of SIGINT
int g_rand_seed = 0;					// random seed

int g_icmp_id = DEFAULT_ICMP_PORT;
int g_udp_id = DEFAULT_UDP_SPORT;
int g_tcp_id = DEFAULT_TCP_DPORT;

int g_curr_id = g_icmp_id;

struct sockaddr_in		g_sin;

char *g_device = "eth0";
int g_random_seed = 0;
int g_probe_type = PROBE_ICMP_ECHO;

char *g_output = NULL;		// output filename
FILE *g_output_f = NULL;		// output file descriptor
char *g_snd_output = NULL;		// output filename
FILE *g_snd_output_f = NULL;		// output file descriptor


eth_t *g_libdnet_eth = NULL;

typedef std::vector<char *> HostV;
HostV g_vhosts;

struct timeval g_recv_end_time;
int     g_receive_mode = 0;             // 0->ICMP socket; 1->TCPDUMP

char *g_src_mac = NULL; // MAC address of source
char *g_dst_mac = NULL; // MAC address of gateway or destination if it is in the same subnet

int g_poisson_ping = 0;	// whether use poisson ping

char *g_recv_probes;		// Received response from the router

char *g_dest_host = NULL;
int g_ttl = 0;

/**********************************************************
 *	                                         Functions
 **********************************************************/

void usage()
{
//	printf("ZIPFrag [-s src_ip] [-i interval] [-I frag_interval] [-c count] [-t type] [-o output] [-p icmp_id] [-r node_list_file | node_list] \n");
	printf("Usage: afad [OPTION]... [ROUTER]...\n");
	printf("Send fragmented probes to destination(s) with poisson interval\n");
	printf("  -s\tsource IP address of the probes\n");
	printf("  -i\taverage probe interval\n");
	printf("  -I\tinterval between two fragments of a datagram\n");
	printf("  -H\tthe destination of the path\n");
	printf("  -l\tlength of probe\n");
	printf("  -c\tnumber of probes\n");	
	printf("  -t\tprobe type: 0->ECHO, 1->Timestamp, 3->TCP\n");
	printf("  -o\toutput file\n");
	printf("  -p\tstarting id of ICMP probes\n");
	printf("  -r\tTTL value\n");
	printf("  -m\tsource MAC address\n");
	printf("  -M\tgateway MAC address\n");
	printf("  -V\tname of the output ethernet interface\n");
	printf("\n");	
	exit(1);
}

// handles SIGINT and SIGSTOP signals 
static void sig(int signo)
{
	switch(signo) {
	case SIGTSTP:
	case SIGINT:
		//printf("Get signal SIGINT(%d)\n", signo);
		if(g_int_state == 0)
			g_int_state = 1;
		else
			exit(1);
		break;
	default:
	        fprintf(stderr, "Error: This signal(%d) shouldn't be caught by me\n", signo);
	}
}

// get exponential distributed random variable
int get_exp_dist_delta_time(int InterArrMean)
{
	double RNTemp, RInterval;
	//int InterArrMean = intervalM * 1000000;

	RNTemp = rand() / (RAND_MAX * 1.0);

	if (RNTemp == 0)
	{
	  	RNTemp = 0.00000001;
	}

	RInterval = (- log(RNTemp) * (InterArrMean - 10.0)) + 10;

	if (RInterval <= 1)
	{
	  	RInterval = 1;
	}

	if (RInterval > (InterArrMean * 10))
	{
	  	RInterval = InterArrMean * 10;
	}

	return (int)(RInterval);
}

double get_exp_dist_delta_time_d(double intervalM)	
{
	int InterArrMean = (int)(intervalM * 1000000);
	return get_exp_dist_delta_time(InterArrMean)/1000000.0;
}

int check_for_reply(int sock, struct sockaddr_in *fromp, struct timeval &wait )
{
	fd_set fds;
	struct timezone tz;
	int cc = 0;
	int fromlen = sizeof(*fromp);

	FD_ZERO(&fds);
	FD_SET(sock, &fds);

	if(g_receive_mode == 0) {
	//	Use raw socket. Can only receive ICMP packet
		if (select(sock + 1, &fds, NULL, NULL, &wait) > 0)
			cc = recvfrom(sock, (char *)g_rcv_packet, sizeof(g_rcv_packet), 0, (struct sockaddr *)fromp, (socklen_t *)&fromlen);
	}
	else {
	//	Use libpcap
		struct ip *ip = (struct ip *) g_rcv_packet;
		cc = getnext_pcap();
		fromp->sin_addr.s_addr = ip->ip_src.s_addr;
	}
	return(cc);
}

int check_packet(int probetype, u_char *buf, int cc, struct sockaddr_in *from, struct sockaddr_in *sin, 
					struct sockaddr_in *din, int &p_id, int &p_seq)
{
	IP_HDR *ip;
	//ICMP_HDR *icp;
	struct icmp *icp;
	TCP_HDR *tcp;		
	u_char type, code;

	// Headers included in an ICMP packet
	IP_HDR *hip;
	UDP_HDR *hup;
	TCP_HDR *htp;		
	struct icmp *hicmp;
	int hlen;
	
	p_id = -1, p_seq = -1;

	ip = (IP_HDR *) buf;
	hlen = ip->ip_hl << 2;
	from->sin_addr.s_addr = ip->ip_src;
	if (cc < hlen + ICMP_MINLEN) {
		fprintf(stderr, "packet too short (%d bytes) from %s\n", cc, inet_ntoa(from->sin_addr));
		return (0);
	}
	
	cc -= hlen;

	// If data is ICMP packet
	//icp = (ICMP_HDR *)(buf + hlen);
	icp = (struct icmp *)(buf + hlen);
	type = icp->icmp_type;
	code = icp->icmp_code;	
	
	if( ip->ip_p == IPPROTO_ICMP) {	
		// Check checksum
		if (in_cksum((u_short *)icp, htons(ip->ip_len) - hlen)) {
			fprintf(stderr, "Packet from: %s: Icmp checksum is wrong\n", inet_ntoa(from->sin_addr));
		/*	for(int x=0; x<20; x++) {
				printf("%x ", *(buf+x));
				if(x%8 == 7)
					printf("\n");
			}
		*/	
			return -1;
		}
	}
	
	if(probetype == PROBE_ICMP_ECHO || probetype == PROBE_ICMP_TIMESTAMP) {
		// Check if this is a right packet		
		//printf("probetype: %d, type: %d, code: %d\n", probetype, type, code);
		if(ip->ip_p != IPPROTO_ICMP)
			return -1;
		
		// If reply is ICMP TIMEXCEED, then TTL has some problem	
		if (type == ICMP_TIMEXCEED && code == ICMP_TIMEXCEED_INTRANS) {
			hip = (IP_HDR *)&icp->icmp_ip;
			hlen = hip->ip_hl << 2;	
			hicmp = (struct icmp *)((u_char *)hip + hlen);

			p_id = ntohs(hicmp->icmp_id);
			p_seq = ntohs(hicmp->icmp_seq);
		//	printf("TTL, id: %d, seq: %d, t1: %d, t2: %d\n", p_id, p_seq, hicmp->icmp_id, g_icmp_id);
			
			if(hip->ip_p == IPPROTO_ICMP && hicmp->icmp_type == ICMP_ECHO && 
			  (p_id == g_icmp_id ||p_id == g_icmp_id+1 ) ) {
			  	if(!g_dest_host)
					fprintf(stderr, "TTL expire, problem\n");
				else {
					p_id--;
					g_recv_probes[p_seq] |= 4;
					return 1;
				}
				return 0;
			}			
			else 
				return -1;
		}
		else if( (type == ICMP_ECHOREPLY && probetype == PROBE_ICMP_ECHO) || 
			(type == ICMP_TSTAMPREPLY && probetype == PROBE_ICMP_TIMESTAMP) ) {
			hip = (IP_HDR *)&icp->icmp_ip;
			if(ip->ip_src  != din->sin_addr.s_addr || ip->ip_dst  != sin->sin_addr.s_addr) {
				//printf("Not my msg, ip->ip_src:%x, din->sin_addr.s_addr:%x, ip->ip_dst:%x, sin->sin_addr.s_addr:%x\n", ip->ip_src, din->sin_addr.s_addr, ip->ip_dst, sin->sin_addr.s_addr);
				return -1;
			}

			p_id = ntohs(icp->icmp_id);
			p_seq = ntohs(icp->icmp_seq);

			// To bypass the limited 16bit ICMP seq number
			int *payload_seq = (int *) (icp->icmp_data);
			if(*payload_seq % 65536 != p_seq) 
				fprintf(stderr, "ICMP data replay is abnormal\n");
			p_seq = *payload_seq;
			
		//	if(p_id == g_icmp_id ||p_id == g_icmp_id+1 )
		//		return 1;
			if(p_id == g_icmp_id ||p_id == g_icmp_id+1 ) {
				if(g_recv_probes[p_seq] & (1<<(p_id-g_icmp_id))) {
					fprintf(stderr, "Duplicate packet: id: %d seq: %d\n", p_id, p_seq);
				}
				if(p_id == g_icmp_id)
					g_recv_probes[p_seq] |= 1;
				else
					g_recv_probes[p_seq] |= 2;
				return 1;
			}
			else
				return -1;
		}		

		return -1;
	}
	else if(probetype == PROBE_TCP_ACK) {
	//	printf("ip->ip_p: %d, IPPROTO_TCP: %d\n", ip->ip_p, IPPROTO_TCP);

		if(ip->ip_p != IPPROTO_TCP || ip->ip_dst  != sin->sin_addr.s_addr)
			return -1;
		
		tcp = (TCP_HDR*) (buf + hlen);
	//	printf("sport: %d, dport: %d\n\n", ntohs(tcp->th_sport), ntohs(tcp->th_dport));
		if( ntohs(tcp->th_sport) != g_tcp_id && ntohs(tcp->th_sport) != g_tcp_id+1) {
			//printf("g_tcp_id:%d, tcp->th_dport:%d\n", g_tcp_id, ntohs(tcp->th_dport));
			return -1;
		}
		if( tcp->th_flags & TH_RST ) {
			p_id = ntohs(tcp->th_sport);
			p_seq = ntohs(tcp->th_dport) - DEFAULT_TCP_SPORT;
		//	p_seq %= MAX_PROBES;
		//	printf("tcp: p_id:%d, p_seq:%d\n", p_id, p_seq);
			return 1;
		}
		//else
		//	printf("tcp->th_flags:%x, TH_RST:%x\n", tcp->th_flags, TH_RST);
		return -1;		
	}
	else {		
		//if (type == ICMP_TIMXCEED && code == ICMP_TIMXCEED_INTRANS) {
		if (type == ICMP_TIMEXCEED && code == ICMP_TIMEXCEED_INTRANS) {
			IP_HDR *hip;
			UDP_HDR *up;
			TCP_HDR *tp;		
			struct icmp *hicmp;
			//ICMP_HDR *hicmp;

			//hip = &icp->icmp_ip;
			hip = (IP_HDR *)&icp->icmp_ip;
			hlen = hip->ip_hl << 2;

		//	if(hip->ip_src.s_addr  != sin->sin_addr.s_addr || hip->ip_dst.s_addr  != din->sin_addr.s_addr)
		//		return -1;
			if(hip->ip_src  != sin->sin_addr.s_addr || hip->ip_dst  != din->sin_addr.s_addr)
				return -1;
				
			if(hlen + 8 <= cc) {	// ICMP TIME EXCEED contains the IP header + 8 bytes data
			//	printf("%d %d %d", hip->ip_p, hip->ip_id, hip->ip_len);
				if( hip->ip_p == IPPROTO_ICMP ) {
					//hicmp = (ICMP_HDR *)((u_char *)hip + hlen);
					hicmp = (struct icmp *)((u_char *)hip + hlen);
					
					p_id = ntohs(hicmp->icmp_id);
					p_seq = ntohs(hicmp->icmp_seq);	
					if(p_id != g_icmp_id)
						return -1;
				//	printf("icmp: p_id: %d, p_seq:%d, hlen:%d\n", p_id, p_seq, hlen);
				}
				else if( hip->ip_p == IPPROTO_UDP ) {
					up = (UDP_HDR *)((u_char *)hip + hlen);
					p_id = ntohs(up->uh_sport);
					p_seq = ntohs(up->uh_dport);
					if(p_id != g_udp_id)
						return -1;
				//	printf("udp: p_id: %d, p_seq:%d, hlen:%d\n", p_id, p_seq, hlen);
				}
				else if( hip->ip_p == IPPROTO_TCP ) {
					tp = (TCP_HDR *)((u_char *)hip + hlen);
					//p_id = ntohs(tp->source);
					//p_seq = ntohl(tp->seq);
					p_id = ntohs(tp->th_sport);
					p_seq = ntohl(tp->th_seq);
					if(p_id != g_tcp_id)
						return -1;				
				//	printf("tcp: p_id: %d, p_seq:%d, hlen:%d\n", p_id, p_seq, hlen);
				}
			}
			else {
				return -2;		// Packet not complete
			}
		}
		else {
			return -1;			// Packet type doesn't match
		}
	}
	return 0;
}


bool gethostaddr(char *hostname, struct sockaddr_in *sin)
{
	int n;
	struct hostent *hp;
	char *p;
	u_int32_t addr, *ap;

	if (strlen(hostname) > 64) {
		fprintf(stderr, "%s: hostname \"%.32s...\" is too long\n", g_prog, hostname);
		return false;
	}
		
	addr = inet_addr(hostname);
	if ((int32_t)addr != -1) {		
		sin->sin_addr.s_addr = addr;
		return true;
	}
	
	hp = gethostbyname(hostname);
	if (hp == NULL) {
		fprintf(stderr, "%s: unknown host %s\n", g_prog, hostname);
		return false;
	}
	if (hp->h_addrtype != AF_INET || hp->h_length != 4) {
		fprintf(stderr, "%s: bad host %s\n", g_prog, hostname);
		return false;
	}

	sin->sin_family = AF_INET;
	ap = &sin->sin_addr.s_addr;
	memcpy((char *)ap,  hp->h_addr_list[0], sizeof(*ap));
	return true;

}

void testsocket()
{
	int sockfd;
	struct sockaddr_in dest;
	char buffer[] = "Hello World!";

	/* create socket , same as client */
	sockfd = socket(PF_INET, SOCK_STREAM, 0);

	/* initialize structure dest */
	bzero(&dest, sizeof(dest));
	dest.sin_family = PF_INET;
	dest.sin_port = htons(8888);
	/* this line is different from client */
	dest.sin_addr.s_addr = INADDR_ANY;

	/* Assign a port number to socket */
	if( bind(sockfd, (struct sockaddr*)&dest, sizeof(dest)) == -1 ) {
		fprintf(stderr, "tcp bind: %s\n", strerror(errno));
		exit (1);
	}

	/* make it listen to socket with max 20 connections */
	listen(sockfd, 20);
	printf("test ok\n");
}

bool init_socket()
{
	int optval = 1;
	struct sockaddr_in sin;


	int sockfd;
	struct sockaddr_in dest;
	char buffer[] = "Hello World!";

	if ((g_rcv_tcp_sock = socket(PF_INET, SOCK_RAW, IPPROTO_TCP)) < 0) {
	//if ((g_rcv_tcp_sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
		fprintf(stderr, "%s: tcp socket: %s\n", g_prog, strerror(errno));
		exit(1);
	}

	optval = 1;
	if (setsockopt(g_rcv_tcp_sock, IPPROTO_IP, IP_HDRINCL, (char *)&optval, sizeof(optval)) < 0) {
		fprintf(stderr, "%s: tcp socket IP_HDRINCL: %s\n", g_prog, strerror(errno));
		exit(-1);
	}

/*
	bzero(&sin, sizeof(sin));
	sin.sin_family = PF_INET;
	sin.sin_port = htons(8888);
	sin.sin_addr.s_addr = INADDR_ANY;
	if( bind(g_rcv_tcp_sock, (struct sockaddr*)&sin, sizeof(sin)) == -1 ) {
		fprintf(stderr, "tcp bind: %s\n", strerror(errno));
		exit (1);
	}
	
	if( listen(g_rcv_tcp_sock, 10) == -1 ) {
		fprintf(stderr, "tcp listen: %s\n", strerror(errno));
		exit (1);
	}
	printf("test ok\n");
*/

	/* Open a safe raw ICMP socket for receiving ICMP errors after sending ICMP echo */
	if ((g_rcv_icmp_sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
		fprintf(stderr, "%s: icmp socket: %s\n", g_prog, strerror(errno));
		exit(-1);
	}

	optval = 1;
	if (setsockopt(g_rcv_icmp_sock, IPPROTO_IP, IP_HDRINCL, (char *)&optval, sizeof(optval)) < 0) {
		fprintf(stderr, "%s: icmp socket IP_HDRINCL: %s\n", g_prog, strerror(errno));
		exit(-1);
       }

//	if( !init_pcap() ) {
//		fprintf(stderr, "%s:init_pcap fails\n", g_prog);
//		exit(-1);
//	}
	
	return true;
}

bool init()
{
//	bool bret = init_socket();	
	bool bret = true;	
	if( !init_pcap() ) {
		fprintf(stderr, "%s:init_pcap fails\n", g_prog);
		exit(-1);
	}
	
	return bret;
}

void random_ip(char *ip)
{
	int i;
	int ipb1, ipb2, ipb3, ipb4;
	while(1) {
		ipb1 = rand() % 224;
		if(ipb1 == 10)
			continue;
		else {
			ipb2 = rand() % 256;
			if(ipb1 == 192 && ipb2 == 168)
				continue;

			ipb3 = rand() % 256;
			ipb4 = rand() % 254 + 1;
			break;
		}
	}
	sprintf(ip, "%d.%d.%d.%d", ipb1, ipb2, ipb3, ipb4);
}

// Receive and check replied packet while waiting between the probing interval
void recv_while_probing(int probetype, int usecs, struct timeval now, int sock, struct sockaddr_in *sin, struct sockaddr_in *din)
{
	struct timeval tp, twait,rt;
	struct timezone tz;
	int p_id, p_seq;
	struct sockaddr_in fromin;
	int cc;	// received packet length

//printf("In recv_while_probing\n");

	// Old approach. Cannot compensate the longer wait for the previous receiving
	/*
	//gettimeofday(&now, &tz);
	twait.tv_sec = usecs/1000000;
	twait.tv_usec = usecs%1000000;
	tp = now;
	tvadd(&tp, &twait);
	*/
	// Use a global varaiable to store time
	if( g_recv_end_time.tv_sec == 0) {
		g_recv_end_time = now;
	}
	twait.tv_sec = usecs/1000000;
	twait.tv_usec = usecs%1000000;
	tvadd(&g_recv_end_time, &twait);
	tp = g_recv_end_time;
	
	do {
		if( (cc = check_for_reply(sock, &fromin, twait)) > 0 ) {
			if( check_packet(probetype, (u_char *)g_rcv_packet, cc, &fromin, sin, din, p_id, p_seq) > 0 ) {
				gettimeofday(&rt, &tz);
				if(!g_output_f) 
					printf("%d.%06ld\t\tR\t%s\t%d\t%d\n", rt.tv_sec, rt.tv_usec, inet_ntoa(fromin.sin_addr),  p_id, p_seq);			
				else
					fprintf(g_output_f, "%d.%06ld\t\tR\t%s\t%d\t%d\n", rt.tv_sec, rt.tv_usec, inet_ntoa(fromin.sin_addr),  p_id, p_seq);			
			}
			else {
			//	printf("bad packet\n");
			}
		}
		/*
		if( (cc = check_for_reply(g_rcv_tcp_sock, &fromin, twait)) > 0 ) {
			if( check_packet(PROBE_TCP_ACK, (u_char *)g_rcv_packet, cc, &fromin, sin, din, p_id, p_seq) > 0 ) {
				printf("p_id: %d, p_seq: %d\n", p_id, p_seq);			
			}
			else {
			//	printf("bad packet\n");
			}
		}
		*/
		usleep(1);
		twait = tp;
		gettimeofday(&now, &tz);						
		tvsub(&twait, &now);
	} while(twait.tv_sec >= 0);
	
//printf("exit recv_while_probing\n");	
}

void log_send(RawDNet *prp)
{
	struct timeval now;
	struct timezone tz;
	int id, seq;
	struct sockaddr_in dst; 
		
	gettimeofday(&now, &tz);
	dst = *prp->get_daddr();
	prp->get_id_and_seq(id, seq);
	if(!g_snd_output_f) 
		printf("%d.%06ld\t\tS\t%s\t%d\t%d\n", now.tv_sec, now.tv_usec, inet_ntoa(dst.sin_addr),  id, seq);			
	else
		fprintf(g_snd_output_f, "%d.%06ld\t\tS\t%s\t%d\t%d\n", now.tv_sec, now.tv_usec, inet_ntoa(dst.sin_addr),  id, seq);			
}

// Generate random integer within [0, max)
int irand(int max)
{
	return (int) (((double)max)*rand()/(RAND_MAX+1.0));
}

bool probe_1router(char *dst, int probetype,  double interval)
{
	RawDNet *prp_whole = NULL, *prp_frags = NULL, *prp_ttl = NULL;
	int i, pktlen;
	int whole_count, frag_count;
	struct sockaddr_in sin, din, hin;
	int sleeptime = (int)(interval * 1e6), frag_sleeptime;
	int sock;
	int psleeptime, tsleeptime = 0;
	struct timeval now, now2;
	struct timezone tz;
	pid_t mypid = getpid();
	PacketChain pchain;
	PacketInChain *pic;

	if( !g_rand_seed )
		srand(getpid());
	else
		srand(g_rand_seed);

	if(g_frag_interval>0)
		frag_sleeptime = (int)(g_frag_interval * 1e6);
	else
		frag_sleeptime = sleeptime;

//	printf("sleeptime: %d, frag_sleeptime: %d\n", sleeptime, frag_sleeptime);
	if( !init_socket() ) {
		return false;
	}
	
	if(probetype == PROBE_ICMP_ECHO || probetype == PROBE_ICMP_TIMESTAMP) {
		prp_whole = new RawDNet(RAW_PACKET_ICMP);
		prp_frags = new RawDNet(RAW_PACKET_ICMP);
		
		if(probetype == PROBE_ICMP_ECHO) {
			prp_whole->get_icmp_itype() = ICMP_ECHO;	
			prp_frags->get_icmp_itype() = ICMP_ECHO;
		}
		else if(probetype == PROBE_ICMP_TIMESTAMP) {
			prp_whole->get_icmp_itype() = ICMP_TSTAMP;	
			prp_frags->get_icmp_itype() = ICMP_TSTAMP;
		}
		
		prp_whole->set_reply_id(g_icmp_id);
		prp_frags->set_reply_id(g_icmp_id+1);
		sock = g_rcv_icmp_sock;		
	}
	else if(probetype == PROBE_TCP_ACK) {
		prp_whole = new RawDNet(RAW_PACKET_TCP);
		prp_frags = new RawDNet(RAW_PACKET_TCP);

		prp_whole->set_tflag(TH_ACK);
		prp_frags->set_tflag(TH_ACK);

		prp_whole->set_reply_id(g_tcp_id);
		prp_frags->set_reply_id(g_tcp_id+1);
		sock = g_rcv_tcp_sock;	
	}
	else {
		assert(0);
	}	
		
	prp_whole->open_eth(g_device);
	prp_frags->open_eth(g_device);
	
//	prp_whole->get_ident() = rand();
//	prp_frags->get_ident() = rand();
	//printf("get rand1: %d, rand2: %d\n", prp_whole->get_ident(), prp_frags->get_ident());

/*
	if( strncmp(dst, "165.124.", 8) != 0 ) {
		prp_whole->set_eth_dst("00:D0:05:2C:98:00");
		prp_frags->set_eth_dst("00:D0:05:2C:98:00");
	}
*/
	if(g_src_mac) {
                prp_whole->set_eth_src(g_src_mac);
                prp_frags->set_eth_src(g_src_mac);
        }
        if(g_dst_mac) {
                prp_whole->set_eth_dst(g_dst_mac);
                prp_frags->set_eth_dst(g_dst_mac);
        }

	gethostaddr(dst, &din);
	prp_whole->set_src(&g_sin);
	prp_whole->set_dst(&din);
	prp_frags->set_src(&g_sin);
	prp_frags->set_dst(&din);
	
	prp_whole->set_pktlen(g_pktlen);
	pktlen = g_pktlen*2 - IP_HEADER_SIZE;
	prp_frags->set_pktlen(pktlen);
	//prp_frags->set_pktlen(g_pktlen+16);
//	prp_whole->prepare_packet();
//	prp_frags->prepare_packet_frag(pktlen, g_pktlen);

	sin = g_sin;

	g_recv_probes = new char[g_count];
	memset(g_recv_probes, 0, sizeof(char)*g_count);
	
	prp_whole->get_ident() = rand();
	prp_frags->get_ident() = rand();
	prp_frags->save_id_seq();

	if(g_dest_host) {		// Prefix path problem. So do TTL-Exceeded probe at the same time
		prp_ttl = new RawDNet(RAW_PACKET_ICMP);
		prp_ttl->get_icmp_itype() = ICMP_ECHO;
		prp_ttl->set_reply_id(g_icmp_id);
		prp_ttl->open_eth(g_device);
		if(g_src_mac) 
			prp_ttl->set_eth_src(g_src_mac);	        
	        if(g_dst_mac) 
	        	prp_ttl->set_eth_dst(g_dst_mac);
	        
		gethostaddr(g_dest_host, &hin);
		prp_ttl->set_src(&g_sin);
		prp_ttl->set_dst(&hin);
		prp_ttl->set_pktlen(g_pktlen);
		prp_ttl->get_ident() = rand();
		prp_ttl->get_ttl() = g_ttl;
	}

//	pchain.add_pkt(0, 0, 0);
	pchain.add_pkt(1, 0, sleeptime/2);
	
	// Sending whole packets and fragmented packets interleaved
	gettimeofday(&now, &tz);
	/*
	// In this code, packets are scattered and may have large measurement error
	while((pic = pchain.head_pkt()) != NULL) {
		if(pic->type() == 0) {		// whole packet
			// send packet
			prp_whole->prepare_packet();
			prp_whole->send_dnet();
			log_send(prp_whole);	
			prp_whole->next_pkt();
		}
		else if(pic->type() <= 2) {
			if(pic->type() == 1) {	// First fragment
				if(pic->seq() < g_count-1) {
					if(g_poisson_ping)
						psleeptime = get_exp_dist_delta_time(sleeptime);	// use poisson ping
					else
						psleeptime = sleeptime;		// do not use poisson ping
					pchain.add_pkt(1, pic->seq()+1, (double)psleeptime);
				}
				//pchain.add_pkt(2, pic->seq(), (double)sleeptime);
				pchain.add_pkt(2, pic->seq(), (double)frag_sleeptime);				
				pchain.add_pkt(0, pic->seq(), (double)sleeptime/2);				
				if(g_dest_host)
					pchain.add_pkt(3, pic->seq(), (double)sleeptime/3);
			}			
			// send out packets
			prp_frags->load_id_seq();
			prp_frags->next_pkt(pic->seq());	
			prp_frags->prepare_packet_frag(pktlen, g_pktlen);
			prp_frags->send_frag(pic->type());
			if(pic->type() == 2)
				log_send(prp_frags);					
		}
		else if(pic->type() == 3) {
			// send packet
			prp_ttl->prepare_packet();
			prp_ttl->send_dnet();
		//	log_send(prp_ttl);	
			prp_ttl->next_pkt();
		}
		gettimeofday(&now, &tz);	
		pic = pchain.pop_pkt();
		recv_while_probing(probetype, (int)(pic->interval()), now, sock, &sin, &din);	
		delete pic;
		pic = NULL;
	}
	*/
	while((pic = pchain.head_pkt()) != NULL) {
		if(pic->type() == 0) {		// whole packet
			// send packet
			prp_whole->prepare_packet();
			prp_whole->send_dnet();
			log_send(prp_whole);	
			prp_whole->next_pkt();
		}
		else if(pic->type() <= 2) {
			if(pic->type() == 1) {	// First fragment
				if(pic->seq() < g_count-1) {
					if(g_poisson_ping)
						psleeptime = get_exp_dist_delta_time(sleeptime);	// use poisson ping
					else
						psleeptime = sleeptime;		// do not use poisson ping
					pchain.add_pkt(1, pic->seq()+1, (double)psleeptime);
				}
				//pchain.add_pkt(2, pic->seq(), (double)sleeptime);
				pchain.add_pkt(2, pic->seq(), (double)frag_sleeptime);				
				pchain.add_pkt(0, pic->seq(), (double)sleeptime/2);//+ irand(1000)-500);		
				if(g_dest_host)
					pchain.add_pkt(3, pic->seq(), (double)sleeptime/2+ irand(1000)-500);
			}			
			// send out packets
			prp_frags->load_id_seq();
			prp_frags->next_pkt(pic->seq());	
			prp_frags->prepare_packet_frag(pktlen, g_pktlen);
			prp_frags->send_frag(pic->type());
			if(pic->type() == 2)
				log_send(prp_frags);					
		}
		else if(pic->type() == 3) {
			// send packet
			prp_ttl->prepare_packet();
			prp_ttl->send_dnet();
		//	log_send(prp_ttl);	
			prp_ttl->next_pkt();
		}		
		pic = pchain.pop_pkt();
		recv_while_probing(probetype, (int)(pic->interval()), now, sock, &sin, &din);	
		delete pic;
		pic = NULL;
		gettimeofday(&now, &tz);	
	}
	
	gettimeofday(&now, &tz);
	recv_while_probing(probetype, 1*1000000, now, sock, &sin, &din);

//	printf("avg receiving time: %g\n", (double)tsleeptime/g_count);
	delete prp_whole;
	delete prp_frags;
	if(prp_ttl)
		delete prp_ttl;
	close(g_rcv_tcp_sock);
	close(g_rcv_icmp_sock);
//	close_pcap();
}

int main(int argc, char *argv[]) 
{
	int sock;
	struct sockaddr_in sin, din, fromin;
//	RunTrace trace;
	int child_pid = 0, status;
	FILE *rtf = NULL, *outf = NULL;	
	
	char *hostname, ipb[64], *routerb, line[256];
	char *cp;
	int op;
	int i,j, ttl = 30, cc;
	int p_id, p_seq;

	struct timeval tp, now, twait, twait2;
	struct timezone tz;
	int pktlen = -1;

	pid_v pids;

	RawDNet *prp;
//	int icmp_count = 0, udp_count = 0, tcp_count = 0; 
	double num1, num2, num3;

	g_recv_end_time.tv_sec = 0;
	g_recv_end_time.tv_usec = 0;

	if (argv[0] == NULL)
		g_prog = "IPFrag";
	else if ((cp = strrchr(argv[0], '/')) != NULL)
		g_prog = cp + 1;
	else
		g_prog = argv[0];

	while ((op = getopt(argc, argv, "ehc:f:i:I:H:l:o:O:p:r:t:s:m:M:V:z:")) != EOF)
		switch (op) {
		case 'c':
			g_count = str2val(optarg, "count", 1, -1);
			break;

		case 'e':
			g_receive_mode = 1;
			break;

		case 'f':
			g_freq = str2val(optarg, "frequency of probing", 1, -1);
			g_interval = 1.0/g_freq;
			break;
			
		case 'i':
			g_interval = atof(optarg);	
			break;

		case 'I':
			g_frag_interval = atof(optarg);	
			break;

		case 'H':
			g_dest_host = optarg;
			break;
			
		case 'l':
			g_pktlen = str2val(optarg, "pack len", 36, 1500);;
			break;

		case 'o':
			g_output = optarg;
			break;
			
		case 'O':
			g_snd_output = optarg;
			break;	
			
  		case 'p':
			g_icmp_id = (u_short)str2val(optarg, "serverport", 0, (1 << 16) - 1);
			if(g_icmp_id == 0) {
				g_icmp_id = getpid();
			}
			g_curr_id = g_icmp_id;
			break;
			
		case 'r':
			g_ttl = str2val(optarg, "TTL", 1, 100);
			break;
			
		case 't':
			g_probe_type = str2val(optarg, "probe type", 0, 6);
		//	if(g_probe_type == PROBE_TCP_ACK)
		//		g_curr_id += 3;
			// *************Other types need support***************
			// *************Not done yet*************************
			break;				
		case 's':
			g_source = optarg;
			break;
		case 'm':
                        g_src_mac = optarg;
                        break;

                case 'M':
                        g_dst_mac = optarg;
                        break;
                
                case 'V':
                        g_device = optarg;
                        break;	
		case 'z':
			g_poisson_ping = str2val(optarg, "ping type", 0, 1);
			break;
		case 'h':
		default:
			usage();
		}

	argc -= optind;
	argv += optind;	

	// shell initialization
//	if (signal(SIGINT, sig) == SIG_ERR) fprintf(stderr, "SIGINT");
//	if (signal(SIGTSTP, sig) == SIG_ERR) fprintf(stderr, "SIGTSTP");

	if(g_probe_type == PROBE_TCP_ACK)
		g_curr_id += 3;
		
	if(g_source == NULL) {
		g_source = (char *)malloc(MAXHOSTNAMELEN+1);
		gethostname(g_source, MAXHOSTNAMELEN);		
	//	printf("%s\n", g_source);
	}
	
	gethostaddr(g_source, &sin);
	g_sin = sin;
			
	if(argc>0) {
		for(int i=0; i<argc; i++) {
			routerb = new char[strlen(argv[i])+1];
			strcpy(routerb, argv[i]);									
			g_vhosts.push_back(routerb);
		}
	}
	else
		usage();

	if(g_dest_host && !g_ttl)
		usage();

	if(!init()) {
		fprintf(stderr, "Fail to initialize\n");
		exit(-1);
	}

	// Open output file if needed
	if(g_output) {
		g_output_f = fopen(g_output, "w");
		if(g_output_f == NULL) {
			printf("Can't open file %s: %s\n", g_output, strerror(errno));
			exit(-1);
		} 
		if(!g_snd_output)
			g_snd_output_f = g_output_f;
	}
	if(g_snd_output) {
		g_snd_output_f = fopen(g_snd_output, "w");
		if(g_snd_output_f == NULL) {
			printf("Can't open file %s: %s\n", g_snd_output, strerror(errno));
			exit(-1);
		}
	}
		
	routerb = g_vhosts[0];
	if( gethostaddr(routerb, &din) ) {
		if(!g_output_f) 
			printf("Probe: %s %d\n", routerb, g_curr_id);
		else
			fprintf(g_output_f, "Probe: %s %d\n", routerb, g_curr_id);
		probe_1router(routerb, g_probe_type, g_interval);
	//	printf("will probe: %s\n", routerb);
	}
	else
		fprintf(stderr, "Bad dest: %s\n", routerb);

	num1 = num2 = num3 = 0;
	for(int i=0; i<g_count; i++) {
		if(g_recv_probes[i] & 1)
			num2++;
		if(g_recv_probes[i] & 2)
			num1++;
		if(g_recv_probes[i] & 4)
			num3++;			
	}

	for(int i=0; i<g_count; i++) {
		printf("%d ", g_recv_probes[i]);
		if(i%50 == 49)
			printf("\n");
	}
	printf("\n");
	
	if(num2>0 && num1>0) {
		if(!g_dest_host) {
			printf("AFAD\tFPING\tPING\n");
			printf("%.4g\t%.4g\t%.4g\n", 1 - num1/num2, 1 - num1/g_count, 1 - num2/g_count);
		//	printf("AFAD loss rate: %g ,fragment loss rate: %g ,whole pkt loss rate: %g\n", 1 - num1/num2, 1 - num1/g_count, 1 - num2/g_count);
		}
		else {
			double fsucc = num1/num2;		// Forward transmission rate measured by AFAD
			double rtsucc = num2/g_count;		// Round-trip transmission rate
			double rsucc = rtsucc / fsucc;		// Reverse path transmission rate
			if(rsucc > 1)
				rsucc = 1;
			double rtsucc_ttl = num3/g_count;	// Round-trip transmission rate by TTL-Exceeded
			double fadsucc = rtsucc_ttl / rsucc;
			printf("AFAD-TTL\tAFAD\tFPING\tPING\tRT-TTL\n");
			printf("%.4g\t%.4g\t%.4g\t%.4g\t%.4g\n", 1 - fadsucc, 1-fsucc, 1 - num1/g_count, 1 - rtsucc, 1 - rtsucc_ttl);
		//	printf("AFAD loss rate: %g , fad forward loss: %g, fragment loss rate: %g ,whole pkt loss rate: %g, ttl loss rate: %g\n", 
		//		1 - fadsucc, 1-fsucc, 1 - num1/g_count, 1 - rtsucc, 1 - rtsucc_ttl);
		}
	}
	else
		printf("# Lose all pings\n");

	for(int i=0; i<g_vhosts.size(); i++)
	{
		delete (char *)g_vhosts[i];
	}
	g_vhosts.clear();
	
	if(g_output_f)
		fclose(g_output_f);
	if(g_snd_output && g_snd_output_f)
		fclose(g_snd_output_f);
	close_pcap();

}


